from typing import List, AnyStr
import math

class BasicChart(object):
    def __init__(self, name="default"):
        self.__name = name

    def _sanitize_list(self, l):
        if hasattr(l, 'tolist'):
            return l.tolist()
        return l

    def set_series_index(self, idx):
        self.__name = self.__name or "Series {}".format(idx)

    @property
    def x_ticks(self):
        return None

    @property
    def y_ticks(self):
        return None

    @property
    def data(self):
        return []

    def as_series(self):
        return {
            "name": self.__name,
            "data": self.data
        }


class Bar(BasicChart):
    def __init__(self, x: List, y: List, name: AnyStr=None, min_val=None, max_val=None, colors=None):
        super(Bar, self).__init__(name=name)
        self.__x = self._sanitize_list(x)
        self.__y = self._sanitize_list(y)
        self._min = min_val
        self._max = max_val
        self._colors = self._validate_colors(colors)

    def _validate_colors(self, colors):
        if colors is None:
            return []
        elif type(colors) is not list:
            raise ValueError("Colors should be in format: ['#3060cf', '#fffbbc', ...]")
        return colors

    @classmethod
    def chart_type(cls):
        return "bar"

    @property
    def x_ticks(self):
        return self.__x

    @property
    def data(self):
        return self.__y

    @property
    def min(self):
        return self._min

    @property
    def max(self):
        return self._max

    @property
    def colors(self):
        return self._colors


class Heatmap(BasicChart):
    def __init__(self, z: List, name: AnyStr=None, color_stops=None, min_val=None, max_val=None):
        super(Heatmap, self).__init__(name=name)
        self._x = []
        self._y = []
        self._z = []
        self._tokenize_list(self._sanitize_list(z))
        self._color_stops = self._validate_stops(color_stops)
        self._min = min_val
        self._max = max_val

    def _validate_stops(self, color_stops):
        if color_stops is None:
            return []
        elif type(color_stops) is not list:
            raise ValueError("Color stops should be in format: [[0, '#3060cf'], [0.5, '#fffbbc'], ...]")

        for stop in color_stops:
            if type(stop) is not list or len(stop) != 2:
                raise ValueError("Color stops should be in format: [[0, '#3060cf'], [0.5, '#fffbbc'], ...]")

        return color_stops

    def _tokenize_list(self, l):
        for q in l:
            if len(q) == 2:
                (x, y), z = q
            elif len(q) == 3:
                x, y, z = q
            else:
                raise ValueError("Heatmap only support in values in type of ((x,y),z) or (x,y,z)")
            if math.isnan(z):
                continue
            if x not in self._x:
                self._x.append(x)
            if y not in self._y:
                self._y.append(y)
            self._z.append((x, y, z))
        self._x = sorted(self._x)
        self._y = sorted(self._y)
        self._z = [(self._x.index(x), self._y.index(y), z) for x, y, z in self._z]

    @classmethod
    def chart_type(cls):
        return "heatmap"

    @property
    def x_ticks(self):
        return self._x

    @property
    def y_ticks(self):
        return self._y

    @property
    def data(self):
        return self._z

    @property
    def color_stops(self):
        return self._color_stops

    @property
    def min(self):
        return self._min

    @property
    def max(self):
        return self._max


class MatrixHeatmap(Heatmap):
    def __init__(self, matrix, name: AnyStr=None, color_stops=None, min_val=None, max_val=None):
        z = []
        for y in range(len(matrix)):
            for x in range(len(matrix[0])):
                z.append((x, y, matrix[y][x]))
        super(MatrixHeatmap, self).__init__(z, name, color_stops, min_val, max_val)


class Scatterplot(BasicChart):
    def __init__(self, x: List, y: List, name: AnyStr=None):
        super(Scatterplot, self).__init__(name=name)
        self._x = self._sanitize_list(x)
        self._y = self._sanitize_list(y)

    @classmethod
    def chart_type(cls):
        return "scatter"

    @property
    def data(self):
        return list(zip(self._x, self._y))


class CategoricalScatterplot(Scatterplot):

    @property
    def x_axis(self):
        return sorted(list(set(self._x)))

    @property
    def y_axis(self):
        return sorted(list(set(self._y)))


    @property
    def data(self):
        x_ax = self.x_axis
        y_ax = self.y_axis
        x_i = list(map(lambda x: x_ax.index(x), self._x))
        y_i = list(map(lambda y: y_ax.index(y), self._y))
        return list(zip(x_i, y_i))





from cnvrg.modules.base_module import CnvrgBase
import cnvrg.helpers.apis_helper as apis_helper
from cnvrg.modules.errors import CnvrgError
import cnvrg.helpers.export_library_helper as export_library_helper
from cnvrg.modules.project import Project
from cnvrg.helpers.param_build_helper import LIBRARY, PROJECT, parse_params
from cnvrg.helpers.url_builder_helper import url_join
from cnvrg.helpers.args_helper import args_to_string, validate_args
from cnvrg.modules.experiment import Experiment
from cnvrg.helpers.param_helper import wrap_string_to_list
from typing import List
import os
import os.path
import sys
DEFAULT_WORKING_DIR = os.path.expanduser("/cnvrg_libraries")
import importlib

class Library(CnvrgBase):
    def __init__(self, library, project=None, working_dir=None, _info=None):
        owner, library = parse_params(library, LIBRARY)
        self.__working_dir = working_dir or DEFAULT_WORKING_DIR
        try:
            self.project = Project(project)
        except CnvrgError as e:
            self.project = None

        self.__library = library
        self.__owner = apis_helper.credentials.owner
        self.__info = _info
        self.__path = None
        self.__cloned = self.__lib_loaded()
        self.model = None
        self.conn = None


    def __base_url(self):
        return url_join("users", self.__owner, "libraries", self.__library)

    @staticmethod
    def export(*paths):
        paths = [os.path.abspath(p) for p in paths]
        resp = export_library_helper.export_library(*paths)
        if resp.get("library"):
            lib_key = resp.get("library").get("key")
            return Library(lib_key)
        else:
            raise CnvrgError(resp.get("message"))

    @staticmethod
    def list():
        owner = apis_helper.credentials.owner
        lib_list = list(map(lambda x: Library(url_join(x.get('owner'), x.get("title")), _info=x), apis_helper.get(url_join("users", owner, "libraries")).get("libraries")))
        return {l.title(): l for l in lib_list}

    def info(self, force=False):
        """
        get info about this library
        :param force: to force api fetching
        :return: dict represent the library
        {clone_cmd: string,
        command: string,
        arguments: list of key: values(list),
        number_of_experiments: integer,
        description: string}
        """
        if self.__info and not force:
            return self.__info
        self.__info = self.__fetch_info()
        return self.info()

    def __fetch_info(self):
        try:
            resp = apis_helper.get(self.__base_url())
            return resp.get("library")
        except Exception as e:
            raise CnvrgError("Cant find library")

    def load(self,load_model=True,install_prerequisites=True):
        """
        load library to your local directory
        :param working_dir: path to clone the library to
        """
        info = self.info()
        lib_dir = self.__lib_path()
        os.makedirs(lib_dir, exist_ok=True)
        export_library_helper.load_library(info.get("package"), lib_dir)
        self.__path = lib_dir
        model_name = info["slug"]
        module_path = lib_dir + "/" + model_name + ".py"
        if load_model and os.path.exists(module_path):
            try:
                spec = importlib.util.spec_from_file_location(model_name,
                                                              module_path)
                mod = importlib.util.module_from_spec(spec)
                spec.loader.exec_module(mod)
                self.model = mod
                if install_prerequisites:
                    self.install_prerequisites()
            except Exception as e:
                # Not always there is a model
                pass

        return lib_dir
    def install_prerequisites(self,pip3=True,bash=True):
        lib_dir = self.__path
        # install requirements
        req_file = lib_dir + "/requirements.txt"
        prerun_file = lib_dir + "/prerun.sh"
        if os.path.exists(req_file):
            if pip3:
                os.system("pip3 install -r %s" % req_file)
            else:
                os.system("pip install -r %s" % req_file)
        if os.path.exists(prerun_file):
            if bash:
                os.system("bash %s" % req_file)
            else:
                os.system("sh %s" % req_file)
    def connect(self, **kwargs):
        if self.model is None:
            raise CnvrgError("Can't find model")
            sys.exit(1)
        self.conn = self.model.connect(**kwargs)
    def query(self, query):
        if self.model is None:
            raise CnvrgError("Can't find model")
            sys.exit(1)
        return self.model.run(conn=self.conn,query=query)
    def to_csv(self, query,filename):
        if self.model is None:
            raise CnvrgError("Can't find model")
            sys.exit(1)
        self.model.to_csv(conn=self.conn,query=query,filename=filename)
    def to_df(self, query,**kwargs):
        if self.model is None:
            raise CnvrgError("Can't find model")
            sys.exit(1)
        return self.model.to_df(conn=self.conn,query=query,**kwargs)
    def to_sql(self, df=None,table_name=None, **kwargs):
        if self.model is None:
            raise CnvrgError("Can't find model")
            sys.exit(1)
        return self.model.to_sql(df=df, conn=self.conn,table_name=table_name,  **kwargs)
    def close_connection(self):
        if self.model is None:
            raise CnvrgError("Can't find model")
            sys.exit(1)
        self.model.close_connection(conn=self.conn)
    def __default_args(self):
        return {
            "project_dir": os.path.abspath(self.project.get_working_dir()),
            "output_dir": os.path.abspath(self.project.get_output_dir())
        }

    def arguments(self):
        return self.info().get("exp_arguments")

    def __mix_args(self, args=None, with_many=True, with_defaults=True):
        args = args or {}
        #exp arguments are key: values
        library_args = self.info().get("exp_arguments") or {}
        all_args = self.__default_args() if with_defaults else {}
        all_args = {
            **all_args,
            **library_args,
            **args
        }
        if not with_many: validate_args(all_args)
        return all_args

    def run(self, compute="local", **kwargs):
        """
        Runs the library on a remote or local compute.
        :param compute: Either a string or a list of computes.
        :param kwargs: All arguments required to run te
        :return: Experiment run
        """
        compute = wrap_string_to_list(compute)
        if "local" in compute:
            return self.__run_local(**kwargs)
        else:
            return self.__run(**kwargs)



    def __run(self, title=None, project=None, compute=None, dataset=None, prerun=True, requirements=None, image=None, sync_before=None, **kwargs):
        ### mix given params with default params
        arguments = self.__mix_args(kwargs)

        ### run experiment.
        return Experiment.run(command=self.info().get("command"), arguments=arguments, library=self.__library,
                              title=title, project=project, compute=compute, dataset=dataset, prerun=prerun, requirements=requirements, image=image, sync_before=sync_before)


    def __run_local(self, title=None, project=None, sync_before=None, sync_after=None, **kwargs):
        #Load the library if needed
        if not self.__lib_loaded(): self.load()

        ### Build the command
        arguments = self.__mix_args(kwargs)
        string_args = args_to_string(arguments)
        command = "{cmd} {args}".format(cmd=self.info().get("command"), args=string_args)

        ## Run a local experiment.
        return Experiment.run(command=command, library=self.__library,  compute='local', local=True, working_directory=self.__lib_path(),
                              sync_before=sync_before, sync_after=sync_after, project=project, title=title)


    def __lib_loaded(self):
        return os.path.exists(os.path.join(self.__lib_path(), "library.yml"))


    def __lib_path(self):
        working_dir = self.__working_dir
        return os.path.join(working_dir, self.info().get("slug"))


    def __str__(self):
        return """{title}
        Description: {description}
        Arguments: {arguments}
        Command: {command}
        """.format(**{**self.info(), "arguments": args_to_string(self.info().get("arguments"))})

    def title(self):
        return url_join(self.info().get("owner"), self.info().get("title"))











from cnvrg.modules.cnvrg_job import CnvrgJob
from cnvrg.modules.project import Project
import cnvrg.helpers.string_helper as string_helper
import cnvrg.helpers.param_build_helper as param_build_helper
from cnvrg.helpers.env_helper import CURRENT_JOB_TYPE, CURRENT_JOB_ID, ENDPOINT, POOL_SIZE
from cnvrg.modules.errors import UserError
from cnvrg.helpers.url_builder_helper import url_join
from cnvrg.helpers.error_catcher import suppress_exception
import cnvrg.helpers.parallel_helper as parallel_helper
import cnvrg.helpers.apis_helper as apis_helper
from cnvrg.modules.dataset import Dataset
from cnvrg.modules.errors import CnvrgError
import json
import sys
import time
import os
import requests
import datetime
import traceback
import pandas as pd
import csv


class Endpoint(CnvrgJob):
    def __init__(self, endpoint=None, project=None):
        owner, project_slug, slug = param_build_helper.parse_params(endpoint, param_build_helper.ENDPOINT)

        p = Project.factory(owner, project)
        if p is None:
            p = Project(url_join(owner, project_slug))

        if not slug:
            raise UserError("Cant create an endpoint without slug")
        slug = slug or CURRENT_JOB_ID
        self.model_name = os.environ.get("CNVRG_MODEL_NAME", "unknown")
        self.log_input = self.should_log_input()
        self.log_output = self.should_log_output()

        self.project = p
        super(Endpoint, self).__init__(slug, ENDPOINT, p)
        self.__data = None

    def __base_log(self, **params):
        log = {
            "job_id": self.job_slug,
            "job_type": self.job_type,
            "owner": self.job_owner_slug,
            "project": self.job_project_slug,
            "model": self.model_name,
            "event_time": time.time(),
            **params
        }
        print(json.dumps(log))
        sys.stdout.flush()

    def log_request(self, input_params="", result="", start_time=None):
        hash = {}
        if self.log_input:
            hash["input"] = str(input_params)[:1000]
        else:
            hash["input"] = ""
        if self.log_output:
            hash["output"] = str(result)[:1000]
        else:
            hash["output"] = ""
        hash["start_time"] = str(datetime.datetime.utcfromtimestamp(int(start_time)))
        end_time = time.time()
        elapsed_time = int((end_time - start_time) * 1000)  # ms
        hash["elapsed_time"] = elapsed_time
        self.__base_log(**hash)

    def log_error(self, e, trace=False):
        hash = {}
        hash["error"] = str(e)
        if trace:
            hash["traceback"] = traceback.format_exc()
        self.__base_log(**hash)

    def log_param(self, tag_name=None, tag_value=None):
        hash = {}
        hash[tag_name] = tag_value
        hash["cnvrg_tag_name"] = tag_name
        self.__base_log(**hash)

    def log_metric(self, name=None, y:float=None, x=None):
        hash = {}
        try:
            hash[name] = float(y)
            hash["y"] = float(y)
        except Exception as e:
            print("ERROR {}".format(e), file=sys.stderr)
            return
        hash["x"] = x
        hash["cnvrg_metric_name"] = name
        self.__base_log(**hash)

    def log_prediction(self, input=None, output=None, elapsed_time=0):
        start_time = datetime.datetime.utcnow().strftime("%c")
        self.__base_log(input=input, output=output, elapsed_time=elapsed_time, start_time=start_time)

    def scale_up(self):
        return apis_helper.post_v2(url_join(self.project.get_base_url(api="v2"), 'endpoints', self.data["slug"], "scale_up"), data={})

    def scale_down(self):
        return apis_helper.post_v2(url_join(self.project.get_base_url(api="v2"), 'endpoints', self.data["slug"], "scale_down"), data={})

    def link_experiment(self):
        job_id = os.environ.get("CNVRG_JOB_ID", "unknown")
        job_type = os.environ.get("CNVRG_JOB_TYPE", "unknown")
        resp = None
        if job_type == "Experiment":
            resp = apis_helper.post_v2(url_join(self.project.get_base_url(api="v2"), 'endpoints', self.data["slug"], "link_experiment"),
                                       data={"experiment_id": job_id})
        return resp

    def is_deployment_running(self):
        resp = apis_helper.get_v2(url_join(self.project.get_base_url(api="v2"), 'endpoints', self.data["slug"], "is_running"))
        return resp.json().get("status")

    def rollback(self):
        resp = apis_helper.post_v2(url_join(self.project.get_base_url(api="v2"), 'endpoints', self.data["slug"], "rollback_model"),
                           data={})
        return resp.json()

    def get_predictions(self, model_title=None):
        url = url_join(self.project.get_base_url(api="v2"), 'endpoints', self.data["slug"], "predictions")
        if model_title:
            url = url_join(url, "?model_title={model_title}".format(model_title=model_title))
        resp = apis_helper.get_v2(url)
        return resp.json()

    def run_batch_predict(self, dataset, input_file, output_file, scale: bool=True):
        if not output_file or not input_file or not dataset:
            raise CnvrgError("output_file, input_file and dataset can\'t be empty")

        try:
            endpoint_type = self.data.get("kind")
            should_scale = endpoint_type == "batch" and scale

            # fetch dataset details
            ds = Dataset(dataset)
            if ds is None:
                raise CnvrgError("Can\'t find Dataset {dataset}".format(dataset=dataset))

            ds_url = ds.get_full_url()

            self.link_experiment()

            if should_scale:
                print("Starting to scale up endpoint")
                self.scale_up()

                is_running = self.is_deployment_running()
                while not is_running:
                    print("Endpoint is not running yet, retrying in 10 seconds")
                    time.sleep(10)
                    is_running = self.is_deployment_running()
                print("Endpoint is online, starting batch prediction")

                time.sleep(20)
            else:
                is_running = self.is_deployment_running()
                if is_running:
                    print("Starting batch prediction")
                else:
                    raise CnvrgError("Endpoint is not running please send scale: True or start your endpoint")

            # Input file should be absolute path
            row_list = []
            data = pd.read_csv(input_file, header=0)
            if not data:
                raise CnvrgError("Input file is empty or invalid")
            for row in data.values:
                try:
                    r_list = row.tolist()
                    resp = self.predict(r_list)
                    row_list.append([r_list, resp.get("prediction")])
                except Exception as e:
                    raise CnvrgError(e)

            # create output file tree if not exists
            dirname = os.path.dirname(output_file)
            if dirname:
                os.makedirs(dirname, exist_ok=True)

            # Output file should be absolute path in /cnvrg
            with open(output_file, 'w') as file:
                writer = csv.writer(file)
                writer.writerow(["input", "prediction"])
                for row in row_list:
                    writer.writerow(row)

            print('Uploading {output_file} to dataset {dataset}'.format(output_file=output_file, dataset=dataset))
            os.system('cnvrg data put {url} {exported_file}'.format(url=ds_url, exported_file=output_file))

            print("Batch prediction has finished")
            if should_scale:
                print("Scaling down endpoint")
                self.scale_down()
        except Exception as e:
            if should_scale:
                self.scale_down()
            raise e

    @property
    def session(self):
        return self.__session()

    @property
    def data(self):
        if self.__data:
            return self.__data
        self._refresh_data()
        return self.__data

    @suppress_exception
    def __get_endpoint(self):
        return apis_helper.get(self.__base_url()).get("endpoint")

    def __base_url(self):
        return url_join(
            self.project.get_base_url(), string_helper.to_snake_case(self.job_type) + "s", self.job_slug
        )

    def _refresh_data(self):
        self.__data = self.__get_endpoint()

    def __session(self):
        session = requests.session()
        session.headers = {
            **apis_helper.JSON_HEADERS,
            "Cnvrg-Api-Key": self.data.get("api_key")
        }
        return session

    def predict(self, o):
        endpoint_url = self.data.get("endpoint_url")
        if not endpoint_url:
            self._refresh_data()
            endpoint_url = self.data.get("endpoint_url")
        return self.session.post(endpoint_url, data=json.dumps({"input_params": o})).json()

    def single_batch_predict(self, o):
        prediction = self.predict(o)
        if isinstance(o, dict):
            return {**o, **prediction}
        else:
            return {"query": o, **prediction}

    def disable_log_input(self):
        self.log_input = False
        os.environ["CNVRG_ENDPOINT_LOG_INPUT"] = "false"

    def should_log_input(self):
        return os.environ.get("CNVRG_ENDPOINT_LOG_INPUT", "").lower() != "false"

    def disable_log_output(self):
        self.log_output = False
        os.environ["CNVRG_ENDPOINT_LOG_OUTPUT"] = "false"

    def should_log_output(self):
        return os.environ.get("CNVRG_ENDPOINT_LOG_OUTPUT", "").lower() != "false"



from functools import wraps
import time


class limit(object):
    class Exception(BaseException):
        pass

    def __init__(self, per_second=None, per_minute=None):
        self.period = 0.0
        self.max_calls = 0
        if per_second:
            self.period = 1.0
            self.max_calls = per_second
        elif per_minute:
            self.period = 60.0
            self.max_calls = per_minute
        else:
            raise limit.Exception("You must provide either per_second,"
                                  "per_minute or per_hour values.")

    calls_counter = 0
    last_call_time = None

    def __call__(self, f):
        def wrapped(*args, **kwargs):
            now = time.time()
            delay = 0.0
            if self.last_call_time is not None:
                timedelta = now - self.last_call_time
                if timedelta <= self.period and self.calls_counter >= self.max_calls:
                    self.calls_counter = 0
                    delay = abs(self.period - timedelta)
            time.sleep(delay)
            self.last_call_time = time.time()
            f(*args, **kwargs)
            self.calls_counter += 1
        return wrapped


from cnvrg.modules.library import Library
from cnvrg.modules.library import Library as AILibrary
from cnvrg.modules.project import Project
from cnvrg.modules.experiment import Experiment
from cnvrg.modules.endpoint import Endpoint
from cnvrg.modules.dataset import Dataset
from cnvrg.modules.flow import Flow
from cnvrg.modules.flow_version import FlowVersion
from cnvrg.modules.storage.file_uploader import FileUploader
from cnvrg.modules.cnvrg_me import Cnvrg
from cnvrg.modules.data_connector import *
from cnvrg.version import VERSION


__version__ = VERSION



from cnvrg.modules.data_connector.base_connector import BaseConnector
from threading import Thread
class BaseStreamConnector(BaseConnector):
    def __init__(self, *args, **kwargs):
        ## topic => [callbacks]
        self.subs = {}
        self._producer = None
        ## topic => Proccess
        self.proccess = {}
        super(BaseStreamConnector, self).__init__(*args, **kwargs)

    def producer(self):
        raise Exception("Not Implemented")

    def consumer(self, *topics, **kwargs):
        raise Exception("Not Implemented")


    def __init_sub(self, topic, **kwargs):
        def __subscription(stream_connector, topic, **kwargs):
            _consumer = stream_connector.consumer(topic, **kwargs)
            for msg in _consumer:
                for cb in stream_connector.subs.get(topic):
                    cb(msg)

        t = Thread(target=__subscription, args=(self, topic), kwargs=kwargs)
        t.start()
        self.proccess[topic] = t
        return t

    def subscribe(self, topic, callback, **kwargs):
        if topic not in self.subs:
            self.__init_sub(topic, **kwargs)
            self.subs[topic] = []
        self.subs[topic].append(callback)
        return self.proccess[topic]


    def unsubscribe(self, topic):
        if topic not in self.proccess:
            return
        self.proccess[topic].terminate()
        self.subs[topic] = None

    @property
    def __producer(self):

        if not self._producer: self._producer = self.producer()
        return self._producer


    def send(self, topic, **kwargs):
        """
        Send a specific **kwargs to a topic
        :param topic: topic to send to
        :return: Future
        """
        return self.__producer.send(topic, **kwargs)


import cnvrg.helpers.apis_helper as apis_helpers
import cnvrg.helpers.config_helper as config_helper
import cnvrg.helpers.env_helper as env_helper
from cnvrg.modules import CnvrgError
import os
## 1
ORGANIZATION = 'ORG'
## 2
PROJECT = 'PRO'
LIBRARY = 'LIBR'
DATASET = 'DATA'
DATA_CONNECTOR = "DATACONNECTOR"
## 3
EXPERIMENT = 'EXP'
NOTEBOOK = 'NOTE'
ENDPOINT = 'ENDP'
FLOW = 'FLOW'
## 4
TAG = 'TAG'
DEPLOYMENT = 'DEPL'


def type_to_depth(d_type):
    if d_type in [ORGANIZATION]: return 1
    if d_type in [PROJECT, DATASET, LIBRARY, DATA_CONNECTOR]: return 2
    if d_type in [EXPERIMENT, NOTEBOOK, ENDPOINT, FLOW]: return 3
    if d_type in [TAG, DEPLOYMENT]: return 4


def min_params(d_type):
    if d_type in [LIBRARY]: return 1
    return 0

def parse_params(params, type=None, working_dir=None):
    working_dir = working_dir or os.curdir
    num_of_params = len(params.split("/")) if params else 0
    if num_of_params < min_params(type):
        raise CnvrgError("Expecting {min_params} params".format(min_params=min_params(type)))
    splitted = params.split("/") if params else []
    if num_of_params == type_to_depth(type):
        return [*splitted]
    if num_of_params == type_to_depth(type) - 1:
        ## assuming logged in
        return [apis_helpers.credentials.owner, *splitted]
    if num_of_params == type_to_depth(type) - 2:
        ### assuming in project
        element = config_helper.get_element_slug(working_dir)
        owner = config_helper.get_element_owner(working_dir)
        return [owner, element, *splitted]
    if num_of_params == type_to_depth(type) - 3:
        current_job_slug = env_helper.get_current_job_id()
        project_slug = config_helper.get_element_slug(working_dir)
        owner = config_helper.get_element_owner(working_dir)
        return [owner, project_slug, current_job_slug]




import os
import math
import json
import time
import hashlib
import fnmatch
import mimetypes
import threading

from queue import Queue
from pathlib import Path
from os import path, walk
from cnvrg.helpers.url_builder_helper import url_join
from cnvrg.modules.storage.progress import Progressbar

import cnvrg.modules.storage as storage
import cnvrg.helpers.apis_helper as apis_helper

class FileUploader:
    def __init__(self, element=None):
        self.element = element
        self.storage_client = storage.storage_factory(element=element)

    def upload_multiple_files(self, fileable, files=None, local_dir=None, commit_sha1=None):
        if files is not None:
            chunks = FileUploader.chunk_file_list(files, math.ceil(len(files)/1000))

            progress = Progressbar(total=len(files))

            progress_bar_lock = threading.Lock()

            files_queue = Queue(5000)
            progress_queue = Queue(5000)

            work = threading.Event()
            work.set()

            files_collector_thread = threading.Thread(
                target=FileUploader.files_collector,
                args=(files_queue, progress, progress_bar_lock, fileable, commit_sha1, chunks, local_dir)
            )
            files_collector_thread.start()

            files_reporter_thread = threading.Thread(
                target=FileUploader.files_reporter,
                args=(progress_queue, progress, progress_bar_lock, fileable, commit_sha1)
            )
            files_reporter_thread.start()

            thread_pool = []
            for i in range(len(chunks)):
                t = threading.Thread(
                    target=FileUploader.worker,
                    args=(work, files_queue, progress_queue, self.storage_client)
                )
                t.start()
                thread_pool.append(t)

            files_collector_thread.join()
            files_reporter_thread.join()

            work.clear()

            for t in thread_pool:
                t.join()

    def download_multiple_files(self, files=None):
        pass
        # if files is not None:
        #     for file in files:
        #         self.storage_client.download_single_file(file, target)

    @staticmethod
    def files_collector(file_queue, progress, progress_bar_lock, fileable, commit_sha1, chunks, local_dir):
        for chunk in chunks:
            try:
                blob_vs = FileUploader.create_blob_versions(fileable, commit_sha1, chunk, local_dir)
                for k, v in blob_vs["files"].items():
                    try:
                        file_queue.put({
                            "full_path": k,
                            "source": v["local_path"],
                            "target": v["path"],
                            "bv_id": str(v["bv_id"])
                        })
                    except Exception:
                        with progress_bar_lock:
                            progress.progress += 1
                        print("Failed to upload {}".format(k))

                with progress_bar_lock:
                    progress.progress += len(chunk) - len(blob_vs["files"])

            except Exception as e:
                with progress_bar_lock:
                    progress.progress += len(chunk)
                print(e)

    @staticmethod
    def files_reporter(progress_queue, progress, progress_bar_lock, fileable, commit_sha1):
        uploaded_files = []

        while progress.progress < progress.total:
            try:
                file = progress_queue.get_nowait()

                with progress_bar_lock:
                    progress.progress += 1

                if file is not None:
                    uploaded_files.append(file["bv_id"])

                    if len(uploaded_files) >= 1000 or progress.progress >= progress.total:
                        FileUploader.connect_blob_versions_to_commit(fileable, uploaded_files, commit_sha1)
                        uploaded_files = []
            except Exception:
                time.sleep(1)

        progress.finish()

    @staticmethod
    def worker(work, files_queue, progress_queue, storage_client):
        while work.is_set():
            try:
                file = files_queue.get_nowait()
                try:
                    storage_client.upload_single_file(file["source"], file["target"])
                    progress_queue.put(file)
                except Exception as e:
                    progress_queue.put(None)
                    print("failed to upload {}".format(file["full_path"]))
                    print(e)

            except Exception as e:
                print(e)
                time.sleep(1)

    @staticmethod
    def chunk_file_list(file_list, num_chunks):
        chunks = []
        for i in range(0, num_chunks):
            chunks.append(file_list[i::num_chunks])
        return chunks

    @staticmethod
    def get_recursive_tree(file_path, is_absolute, is_dir, depth=False):
        trees = []
        if is_absolute:
            if depth:
                while str(file_path) != Path(file_path).anchor and str(file_path) != ".":
                    tmp_path = Path(file_path)
                    tmp_parent = tmp_path.parent
                    trees.append("{}/".format(str(tmp_parent)))
                    file_path = tmp_parent
            else:
                if is_dir and not file_path.endswith("/"):
                    file_path = "{}/".format(file_path)
                trees.append(file_path)
        else:
            while str(file_path) != Path(file_path).anchor and str(file_path) != ".":
                tmp_path = Path(file_path)
                tmp_parent = tmp_path.parent
                trees.append("{}/".format(str(tmp_parent)))
                file_path = tmp_parent
        return trees

    @staticmethod
    def get_files_and_dirs_recursive(root_dir=".", regex="*"):
        full_paths = [url_join(FileUploader.get_relative_path(root_dir), "/")]
        for root, dirs, files in walk(root_dir, topdown=False):
            for name in files:
                full_paths.append(FileUploader.get_relative_path(path.join(root, name)))
            for name in dirs:
                full_paths.append(url_join(FileUploader.get_relative_path(path.join(root, name)), "/"))

        return fnmatch.filter(full_paths, regex)

    @staticmethod
    def get_relative_path(full_path):
        if full_path.startswith("./"):
            full_path = full_path[2:]
        return full_path

    @staticmethod
    def sha1(full_path):
        BUF_SIZE = 65536
        sha1 = hashlib.sha1()

        with open(full_path, 'rb') as f:
            while True:
                data = f.read(BUF_SIZE)
                if not data:
                    break
                sha1.update(data)

        return sha1.hexdigest()

    @staticmethod
    def create_blob_versions(fileable, commit_sha1, full_paths, local_dir):
        owner = fileable.owner
        files = {}
        for full_path in full_paths:
            full_path = full_path.replace("./", "")
            is_file = os.path.isfile(full_path)
            is_directory = os.path.isdir(full_path)
            is_absolute = os.path.isabs(full_path)
            if is_absolute:
                relative_path = os.path.relpath(full_path, local_dir) if local_dir else os.path.basename(full_path)
            else:
                relative_path = full_path

            if is_file:
                content_type = mimetypes.guess_type(full_path, strict=True)
                files.update({
                    full_path: {
                        "absolute_path": "{}/".format(os.path.dirname(os.path.abspath(full_path))),
                        "local_path": full_path,
                        "content_type": content_type[0] or "plain/text",
                        "file_name": os.path.basename(full_path),
                        "file_size": os.path.getsize(full_path),
                        "relative_path": relative_path,
                        "sha1": FileUploader.sha1(full_path)
                    }
                })
            elif is_directory:
                files.update({
                    full_path: {
                        "absolute_path": "{}/".format(url_join(os.getcwd(), full_path)),
                        "relative_path": full_path
                    }
                })

        resp = apis_helper.post_v2(url_join(owner, "files", "create_blob_versions"), data={
            "fileable_type": "Project",
            "fileable_slug": fileable.slug,
            "commit": commit_sha1,
            "files": json.dumps(files)
        })

        return resp.json().get("result")

    @staticmethod
    def connect_blob_versions_to_commit(fileable, blob_ids, commit_sha1):
        owner = fileable.owner
        resp = apis_helper.post_v2(url_join(owner, "files", "connect_blob_versions_to_commit"), data={
            "fileable_type": "Project",
            "fileable_slug": fileable.slug,
            "commit": commit_sha1,
            "blob_ids": blob_ids
        })

        return resp.json()



import os

from typing import Dict
from azure.storage.blob import BlobServiceClient
from cnvrg.modules.cnvrg_files import CnvrgFiles
from cnvrg.modules.storage.base_storage import BaseStorage


class AzureStorage(BaseStorage):
    def __init__(self, element: CnvrgFiles, working_dir: str, storage_resp: Dict):
        super().__init__(element, working_dir, storage_resp.get("sts"))

        try:
            os.remove(self.sts_local_file)
        except Exception:
            pass

        del storage_resp["sts"]

        props = self.decrypt_dict(storage_resp, keys=["container", "storage_access_key", "storage_account_name", "container"])
        account_name = props["storage_account_name"]
        accout_key = props["storage_access_key"]
        container = props["container"]

        self.access_key = "DefaultEndpointsProtocol=https;AccountName={};AccountKey={};EndpointSuffix=core.windows.net".format(account_name, accout_key)
        self.container_name = container
        self.service = self._get_service()

    def upload_single_file(self, file, target):
        try:
            client = self.service.get_blob_client(self.container_name, blob=target)
            with open(file, "rb") as data:
                client.upload_blob(data, overwrite=True)
        except Exception as e:
            print(e)

    def download_single_file(self, file, target):
        pass

    def _get_service(self):
        return BlobServiceClient.from_connection_string(self.access_key)



from .base_module import CnvrgBase
import cnvrg.helpers.config_helper as config_helper
from cnvrg.helpers.hash_helper import hash_dir
from cnvrg.helpers.apis_helper import post as apis_post, get as apis_get, put as apis_put
import cnvrg.helpers.cnvrgignore_helper as cnvrgignore_helper
from cnvrg.helpers.url_builder_helper import url_join
import yaml
import os

class CnvrgFiles(CnvrgBase):
    def __init__(self, in_dir=False):
        self.__config = {}
        self.__working_dir = None
        self.in_dir = in_dir
        if in_dir: self.__config = config_helper.load_config(".")


    ### Need to be implemented
    def _default_config(self) -> str:
        pass

    def get_working_dir(self) -> str:
        return self.__working_dir

    def get_base_url(self) -> str:
        pass

    def __save_config(self):
        self.__config = self.__config or self._default_config()
        config_helper.save_config(self.__config, self.get_working_dir())

    def __tree_path(self):
        return os.path.join(self.get_working_dir(), '.cnvrg', 'tree.yml')


    def get_current_commit(self):
        return config_helper.load_idx(self.__working_dir).get(":commit")

    def get_current_tree(self):
        return hash_dir(self.get_working_dir())


    def _generate_tree(self):
        tree = self.get_current_tree()
        os.makedirs(os.path.dirname(self.__tree_path()), exist_ok=True)
        with open(self.__tree_path(), 'w') as f:
            yaml.dump(tree, f)

    def set_commit(self, commit_sha1):
        self.__config["commit"] = commit_sha1
        self.__save_config()


    def _set_working_dir(self, working_dir, save=False):
        self.__working_dir = working_dir
        if save: self.__save_config()



    def get_latest_tree(self):
        if not os.path.exists(self.__tree_path()):
            return {}
        os.makedirs(os.path.dirname(self.__tree_path()), exist_ok=True)
        with open(self.__tree_path(), 'r') as f:
            return yaml.safe_load(f)


    def get_storage_url(self):
        return url_join(self.get_base_url(), "client")


    def fetch_commit(self, commit_sha1=None, limit=10000, offset=0, files=None):
        commit_sha1 = commit_sha1 or "latest"
        commit = apis_get(url_join(self.get_base_url(), "commits", commit_sha1), data={"limit": limit, "offset": offset, "files": files}).get("commit")
        return commit

    def get_commit_files(self, **kwargs):
        commit_size = None
        current_offset = 0
        blobs = []
        while current_offset != commit_size:
            kwargs["offset"] = current_offset
            resp = self.fetch_commit(**kwargs)
            blobs += resp.get("blobs")
            commit_size = resp.get("files_count")
            current_offset = len(blobs)
        return blobs


    def start_commit(self, is_branch=False, force=False) -> str:
        """
        :param kwargs: force, new_branch
        :return: commit_sha1 for the new commit
        """
        commit = apis_post(url_join(self.get_base_url(), 'commits'), data={"force": force, "is_branch": is_branch, "commit_sha1": self.get_current_commit()}).get("commit")
        return commit.get("sha1")

    def end_commit(self, sha1) -> bool:
        """
        :param sha1: the sha1 of the commit we want to finish.
        :return: saved successfuly, true/false
        """
        success = apis_post(url_join(self.get_base_url(), 'commits', sha1)).get("status") == 200
        if not success: return False
        self.set_commit(sha1)
        self._generate_tree()
        self.__save_config()
        return True


    def status(self, current_tree=None, commit=None, to_commit=None, force=False, new_branch=False):
        """
        Current project Status
        :param current_tree: current tree
        :param commit: current commit
        :param to_commit: commit to downlaod
        :param force: do we force our upload
        :param new_branch: we are uploading to a new branch
        :return: {updated_on_server, updated_on_local, conflicts, deleted_conflicts, deleted_on_local, deleted_on_remote}
        """
        commit = commit or self.get_current_commit()
        current_tree = current_tree or self.get_current_tree()
        resp = apis_post(url_join(self.get_base_url(), "commits", "status"), data={"tree": current_tree, "commit": commit, "to_commit": to_commit, "ignore": list(cnvrgignore_helper.get_full_ignore(self.get_working_dir()))})
        status = resp.get("status")
        if force or new_branch:
            status["conflicts"] = []
            status["deleted_conflicts"] = []
        return status

    def upload_files(self, commit_sha1, upload_payload):
        """
        :param upload_payload: {added_files, deleted_files}
        :return: added_files mapping
        """
        return apis_put(url_join(self.get_base_url(), 'commits', commit_sha1), data=upload_payload)




    def _clone_finish(self, commit_sha1, working_dir):
        self._set_working_dir(os.path.realpath(working_dir), save=True)
        self._download_finish(commit_sha1)

    def _download_finish(self, commit_sha1):
        self.set_commit(commit_sha1)
        cnvrgignore_helper.check_cnvrgignore(self.get_working_dir())
        self._generate_tree()
        self.__save_config()

    def get_config(self):
        return self.__config

    def set_config(self, **new_config):
        self.__config = {**self.__config, **new_config}






from setuptools import setup, find_packages
from cnvrg.version import VERSION
setup(
    name='cnvrg',
    version=VERSION,
    python_requires='>3.5',
    packages=find_packages(exclude=("tests",)),
    install_requires=[
        'click', 'requests', "boto3", 'colorama', "tqdm", "pyaml", 'tinynetrc', 'pycryptodome', 'psutil', 'dill', 'pytz', 'numpy', 'pandas', 'azure-storage-blob>=12.1.0', 'google-cloud-storage', 'progress'
    ],
    author="cnvrg",
    author_email="support@cnvrg.io",
    entry_points='''
        [console_scripts]
        cnvrgp=cnvrg.main:cli
    ''',
)



def wrap_string_to_list(payload):
    if payload == None: return None
    if isinstance(payload, str): payload = [payload]
    return payload

def wrap_into_list(payload):
    if payload == None: return []
    if isinstance(payload, list): return payload
    return [payload]


import os

from typing import Dict, List
from cnvrg.modules import UnknownStsError
from cnvrg.modules.base_module import CnvrgBase
from cnvrg.modules.cnvrg_files import CnvrgFiles
from cnvrg.helpers.crypto_helpers import decrypt
from cnvrg.helpers.apis_helper import download_file
from typing import Dict, List
import os


class BaseStorage(CnvrgBase):
    def __init__(self, element: CnvrgFiles, working_dir: str, sts_path: str):
        self.element = element
        self.working_dir = working_dir

        self.sts_local_file = None
        self.key = None
        self.iv = None
        self.init_sts(sts_path)

        self.conflicts = []

    def init_sts(self, sts_path):
        sts_local_file = os.path.join(os.path.expanduser("~"), ".cnvrg", ".sts")
        sts_file = download_file(sts_path, sts_local_file)

        if not sts_file or not os.path.exists(sts_file):
            raise UnknownStsError("Cant find sts")

        with open(sts_file) as content:
            self.sts_local_file = sts_local_file
            if not content:
                raise UnknownStsError("Cant find sts")
            sts_content = content.read().split("\n")
            self.key = sts_content[0]
            self.iv = sts_content[1]

    def decrypt(self, text):
        return decrypt(self.key, self.iv, text)

    def decrypt_dict(self, props: Dict, keys: List = None):
        return {k: decrypt(self.key, self.iv, v) if k in keys else v for k, v in props.items()}

